// PointBillboard.geom

// SetParameter GL_GEOMETRY_INPUT_TYPE_EXT=GL_POINTS
// SetParameter GL_GEOMETRY_OUTPUT_TYPE_EXT=GL_TRIANGLE_STRIP
// SetParameter GL_GEOMETRY_VERTICES_OUT_EXT=6

#version 150
#extension GL_EXT_geometry_shader4 : enable
#extension GL_EXT_gpu_shader4 : enable

layout (points) in;
layout (triangle_strip, max_vertices = 6) out;

// ModelMatrix = viewInv * ModelView;
uniform mat4 baseInvModelMatrix;//
uniform mat4 modelInvBaseMatrix;//
uniform mat4 modelViewMatrix;
uniform mat4 modelViewProjectionMatrix;
uniform int twistMode;
uniform int ortho;

in vec3 modelview_positionIn[];
in float pointSizeIn[];
in vec4 colorIn[];
in vec4 colorDiffuseIn[];
in vec4 colorSpecularIn[];
out vec3 modelview_position;
out vec4 modelview_position4;
out vec4 color;
out vec4 colorDiffuse;
out vec4 colorSpecular;
out vec4 normal;
out vec4 tangent;
out vec4 binormal;
out vec2 gsTexCoord;
flat in int blendIn[];
flat in int pseudoSphereIn[];
flat in int textureImageIndexIn[];
flat in int twistIDIn[];

flat out int blend;
flat out int pseudoSphere;
flat out int textureImageIndex;

void main(void){

    int length = gl_PositionIn.length();
    
    if (twistMode == 0){
	for (int i = 0; i < length; i++){

	    vec4 baseCoord = baseInvModelMatrix * gl_PositionIn[i];
	
	    float x = baseCoord[0];
	    float y = baseCoord[1];
	    float z = baseCoord[2];
	
	    float norm = sqrt(x*x+y*y+z*z);
	    if (norm == 0) continue;

	    x /= norm;
	    y /= norm;
	    z /= norm;
	
	    float sqxy = sqrt(x*x+y*y);
       
	    vec4 normalBase = vec4(x, y, z, 0);
	    vec4 tangentBase = -vec4(-y/sqxy, x/sqxy, 0, 0);
	    vec4 binormalBase = -vec4(-x*z/sqxy, -y*z/sqxy, sqxy, 0);

	    if (ortho == 1){
		normalBase   = vec4(0,0,-1, 0);
		tangentBase  = vec4(1,0,0, 0);
		binormalBase = vec4(0,1,0, 0);
	    }

	    normal = modelInvBaseMatrix * normalBase;
	    tangent  = modelInvBaseMatrix * tangentBase;
	    binormal = modelInvBaseMatrix * binormalBase;

	    vec4 offset[4] = vec4[]((-tangent-binormal), (+tangent-binormal),
				    (-tangent+binormal), (+tangent+binormal));
	    vec2 gs[4] = vec2[](vec2(1,0), vec2(0, 0), vec2(1,1), vec2(0,1));

	    normal = modelViewMatrix * normal;
	    tangent = modelViewMatrix * tangent;
	    binormal = modelViewMatrix * binormal;
	    
	    blend = blendIn[i];
	    pseudoSphere = pseudoSphereIn[i];
	    textureImageIndex = textureImageIndexIn[i];
	    modelview_position = modelview_positionIn[i];
	    modelview_position4 = vec4(modelview_positionIn[i], 1);
	    color = colorIn[i];
	    colorDiffuse = colorDiffuseIn[i];
	    colorSpecular = colorSpecularIn[i];

	    float psize = pointSizeIn[i];
	    //psize = 0.1;
	
	    if (color[3] > 0){
		gsTexCoord.st = gs[0];
		gl_Position = modelViewProjectionMatrix * (gl_PositionIn[i] + offset[0] * psize) ;
		modelview_position4 = modelViewMatrix * (gl_PositionIn[i] + offset[0] * psize) ;
		EmitVertex();
	    
		gsTexCoord.st = gs[2];
		gl_Position = modelViewProjectionMatrix *  (gl_PositionIn[i] + offset[2] * psize);
		modelview_position4 = modelViewMatrix * (gl_PositionIn[i] + offset[2] * psize) ;
		EmitVertex();
	    
		gsTexCoord.st = gs[1];
		gl_Position = modelViewProjectionMatrix *  (gl_PositionIn[i] + offset[1] * psize);
		modelview_position4 = modelViewMatrix * (gl_PositionIn[i] + offset[1] * psize) ;
		EmitVertex();
	    
		gsTexCoord.st = gs[2];
		gl_Position = modelViewProjectionMatrix *  (gl_PositionIn[i] + offset[2] * psize);
		modelview_position4 = modelViewMatrix * (gl_PositionIn[i] + offset[2] * psize) ;
		EmitVertex();
	    
		gsTexCoord.st = gs[1];
		gl_Position = modelViewProjectionMatrix *  (gl_PositionIn[i] + offset[1] * psize);
		modelview_position4 = modelViewMatrix * (gl_PositionIn[i] + offset[1] * psize) ;
		EmitVertex();
	    
		gsTexCoord.st = gs[3];
		gl_Position = modelViewProjectionMatrix *  (gl_PositionIn[i] + offset[3] * psize);
		modelview_position4 = modelViewMatrix * (gl_PositionIn[i] + offset[3] * psize); 
		EmitVertex();
	    }
	}
	
	EndPrimitive();
    }
    else{
	float rad0 = -3.14159/4;
	float rad1 = 3.14159/4;
	float rad2 = 3.14159*3/4;
	float rad3 = 3.14159*5/4;

	float radius = sqrt(2)/2;
	for (int i = 0; i < length; i++){
	
	    vec4 baseCoord = baseInvModelMatrix * gl_PositionIn[i];
	
	    float x = baseCoord[0];
	    float y = baseCoord[1];
	    float z = baseCoord[2];
	
	    float norm = sqrt(x*x+y*y+z*z);
	    if (norm == 0) continue;

	    x /= norm;
	    y /= norm;
	    z /= norm;
	
	    float sqxy = sqrt(x*x+y*y);
       
	    vec4 normalBase = vec4(x, y, z, 0);
	    vec4 tangentBase = -vec4(-y/sqxy, x/sqxy, 0, 0);
	    vec4 binormalBase = -vec4(-x*z/sqxy, -y*z/sqxy, sqxy, 0);

	    normal = modelInvBaseMatrix * normalBase;
	    tangent  = modelInvBaseMatrix * tangentBase;
	    binormal = modelInvBaseMatrix * binormalBase;

	    vec4 offset[4] = vec4[]((-tangent-binormal), (+tangent-binormal),
				    (-tangent+binormal), (+tangent+binormal));

	    normal   = modelViewMatrix * normal;
	    tangent  = modelViewMatrix * tangent;
	    binormal = modelViewMatrix * binormal;
	
	    blend = blendIn[i];
	    pseudoSphere = pseudoSphereIn[i];
	    textureImageIndex = textureImageIndexIn[i];
	    modelview_position = modelview_positionIn[i];
	    modelview_position4 = vec4(modelview_positionIn[i], 1);
	    color = colorIn[i];
	    colorDiffuse = colorDiffuseIn[i];
	    colorSpecular = colorSpecularIn[i];
	    float twist = float(twistIDIn[i] % 1000);

	    float psize = pointSizeIn[i];
	    //psize = 0.1;
	
	    if (color[3] > 0){
		gsTexCoord.st = vec2(radius*cos(rad0+twist)+0.5, radius*sin(rad0+twist)+0.5);
		gl_Position = modelViewProjectionMatrix * (gl_PositionIn[i] + offset[0] * psize) ;
		EmitVertex();
	    
		gsTexCoord = vec2(radius*cos(rad1+twist)+0.5, radius*sin(rad1+twist)+0.5);
		gl_Position = modelViewProjectionMatrix *  (gl_PositionIn[i] + offset[2] * psize);
		EmitVertex();
	    
		gsTexCoord = vec2(radius*cos(rad3+twist)+0.5, radius*sin(rad3+twist)+0.5);
		gl_Position = modelViewProjectionMatrix *  (gl_PositionIn[i] + offset[1] * psize);
		EmitVertex();
	    
		gsTexCoord = vec2(radius*cos(rad1+twist)+0.5, radius*sin(rad1+twist)+0.5);
		gl_Position = modelViewProjectionMatrix *  (gl_PositionIn[i] + offset[2] * psize);
		EmitVertex();
	    
		gsTexCoord = vec2(radius*cos(rad3+twist)+0.5, radius*sin(rad3+twist)+0.5);
		gl_Position = modelViewProjectionMatrix *  (gl_PositionIn[i] + offset[1] * psize);
		EmitVertex();
	    
		gsTexCoord = vec2(radius*cos(rad2+twist)+0.5, radius*sin(rad2+twist)+0.5);
		gl_Position = modelViewProjectionMatrix *  (gl_PositionIn[i] + offset[3] * psize);
		EmitVertex();
	    }
	}
	
	EndPrimitive();
    }

    
}
