// StreamlineTriangles.geom

// SetParameter GL_GEOMETRY_INPUT_TYPE_EXT=GL_LINES_ADJACENCY_EXT
// SetParameter GL_GEOMETRY_OUTPUT_TYPE_EXT=GL_TRIANGLE_STRIP
// SetParameter GL_GEOMETRY_VERTICES_OUT_EXT=6

#version 150
//#extension GL_EXT_geometry_shader4 : enable
//#extension GL_EXT_gpu_shader4 : enable

layout (lines_adjacency) in;
layout (triangle_strip, max_vertices = 6) out;

uniform float lineWidth;
uniform vec4 viewport;
uniform int mappingType;

//varying in vec3 positionIn[];
in vec4 vertexIn[];
in vec4 texCoordIn[];
in vec4 colorIn[];
in vec4 positionIn[];
in vec3 modelview_positionIn[];
flat in int lineIndexIn[];
flat in float filterAlphaIn[];
	
out vec4 texCoord;
out vec2 texCoord2;
out vec4 color;
out vec3 position;
out vec3 modelview_position;
out vec3 vertex;
flat out int lineIndex;
flat out float filterAlpha;

void main(void){

    vec4 posPrev = positionIn[0];
    vec4 posA =    positionIn[1];
    vec4 posB =    positionIn[2];
    vec4 posNext = positionIn[3];
        
    vec2 castPrev = vec2
	((posPrev.x / posPrev.w * viewport[2] + viewport[2])/2 + viewport[0],
	 (posPrev.y / posPrev.w * viewport[3] + viewport[3])/2 + viewport[1]);
    vec2 castA = vec2
	((posA.x / posA.w * viewport[2] + viewport[2])/2 + viewport[0],
	 (posA.y / posA.w * viewport[3] + viewport[3])/2 + viewport[1]);
    vec2 castB = vec2
	((posB.x / posB.w * viewport[2] + viewport[2])/2 + viewport[0],
	 (posB.y / posB.w * viewport[3] + viewport[3])/2 + viewport[1]);
    vec2 castNext = vec2
	((posNext.x / posNext.w * viewport[2] + viewport[2])/2 + viewport[0],
	 (posNext.y / posNext.w * viewport[3] + viewport[3])/2 + viewport[1]);
    
    vec2 normalA0 = normalize(castA - castPrev);
    vec2 normalA1 = normalize(castB - castA);
    vec2 normalA = normalize(normalA0 + normalA1);
    if (lineIndexIn[1] == -1) normalA = normalA1;
    vec4 tangentA = vec4(normalA.y, -normalA.x, 0, 0);
    
    vec2 normalB0 = normalize(castB - castA);
    vec2 normalB1 = normalize(castNext - castB);
    vec2 normalB = normalize(normalB0 + normalB1);
    if (lineIndexIn[3] == -1) normalB = normalB0;
    vec4 tangentB = vec4(normalB.y, -normalB.x, 0, 0);
    
    float width = lineWidth * 0.005;

    gl_Position = (posA - 0.5*width*tangentA);
    position = positionIn[1].xyz;
    modelview_position = modelview_positionIn[1];
    vertex = vertexIn[1].xyz;
    lineIndex = lineIndexIn[2];
    filterAlpha = filterAlphaIn[2];
    color = colorIn[1];
    texCoord = texCoordIn[1];
    if (mappingType == 1) texCoord2 = vec2(texCoordIn[1].s, 0);
    else                  texCoord2 = vec2(0, 0);
    EmitVertex();

    gl_Position = (posB - 0.5*width*tangentB);
    position = positionIn[2].xyz;
    modelview_position = modelview_positionIn[2];
    vertex = vertexIn[2].xyz;
    lineIndex = lineIndexIn[2];
    filterAlpha = filterAlphaIn[2];
    color = colorIn[2];
    texCoord = texCoordIn[2];
    if (mappingType == 1) texCoord2 = vec2(texCoordIn[2].s, 0);
    else                  texCoord2 = vec2(1, 0);
    EmitVertex();

    gl_Position = (posA + 0.5*width*tangentA);
    position = positionIn[1].xyz;
    modelview_position = modelview_positionIn[1];
    vertex = vertexIn[1].xyz;
    lineIndex = lineIndexIn[2];
    filterAlpha = filterAlphaIn[2];
    color = colorIn[1];
    texCoord = texCoordIn[1];
    if (mappingType == 1) texCoord2 = vec2(texCoordIn[1].s, 1);
    else                  texCoord2 = vec2(0, 1);
    EmitVertex();
    EndPrimitive();

    gl_Position = (posB - 0.5*width*tangentB);
    position = positionIn[2].xyz;
    modelview_position = modelview_positionIn[2];
    vertex = vertexIn[2].xyz;
    lineIndex = lineIndexIn[2];
    filterAlpha = filterAlphaIn[2];
    color = colorIn[2];
    texCoord = texCoordIn[2];
    if (mappingType == 1) texCoord2 = vec2(texCoordIn[2].s, 0);
    else                  texCoord2 = vec2(1, 0);
    EmitVertex();

    gl_Position = (posB + 0.5*width*tangentB);
    position = positionIn[2].xyz; 
    modelview_position = modelview_positionIn[2];
    vertex = vertexIn[2].xyz;
    lineIndex = lineIndexIn[2];
    filterAlpha = filterAlphaIn[2];
    color = colorIn[2];
    texCoord = texCoordIn[2];
    if (mappingType == 1) texCoord2 = vec2(texCoordIn[2].s, 1);
    else                  texCoord2 = vec2(1, 1);
    EmitVertex();

    gl_Position = (posA + 0.5*width*tangentA);
    position = positionIn[1].xyz; 
    modelview_position = modelview_positionIn[1];
    vertex = vertexIn[1].xyz;
    lineIndex = lineIndexIn[2];
    filterAlpha = filterAlphaIn[2];
    color = colorIn[1];
    texCoord = texCoordIn[1];
    if (mappingType == 1) texCoord2 = vec2(texCoordIn[1].s, 1);
    else                  texCoord2 = vec2(0, 1);
    EmitVertex();
    EndPrimitive();
}
